# Auxiliary programs for Honore-Tamer example
Φ(x) = cdf(Normal(0.0, 1.0), x)

@with_kw mutable struct HTModel
    T::Int64               = 2
    λ::Array{Float64,1}    = vcat(-3:0.2:3)

    nλ::Int64              = length(λ)
    nx::Int64              = 2 * nλ

    X::Array{Float64,2}    = hcat(vcat(zeros(nλ), ones(nλ)), vcat(λ, λ))
    y::Array{Int64,2}      = hcat(digits.(0:2^T-1, base = 2, pad = T)...)
    Y::Array{Int64,2}      = hcat(digits.(0:2^(T+1)-1, base = 2, pad = T+1)...)

    P::Array{Float64,2}    = zeros(2^T, nx)

    A::Array{Float64,2}    = zeros(nx, 2^T)
    b::Array{Float64,1}    = zeros(nx)
end

# Honore-Tamer prior
function ht_prior(mod)

    λ = mod.λ
    Πy = [0.5 0.5]
    λλ = 0.5 * (λ[2:end] + λ[1:end-1])
    ππ = Φ.(λλ)
    Πλ = vcat(ππ[1], ππ[2:end] - ππ[1:end-1], 1.0 - ππ[end])

    return vcat(Πy[1] * Πλ, Πy[2] * Πλ)

end

# Computes dynamic probit probability for a sequence y = (yt, yt-1, ..., y0)
function dynamic_probit(y, β, a)

    p = 1.0
    for i in 1:length(y)-1
        q = Φ(β * y[i+1] + a)
        if y[i] == 1
            p *= q
        else
            p *= 1.0 - q
        end
    end

    return p

end

# Fills matrix of cell probabilities P corresponding to β
# Updates in-place a 2^T x nx matrix of model-implied probabilities for each
# outcome at each point in the support of the prior
function make_cell_probabilities!(mod, β)

    @unpack P, Y, λ, T, nλ = mod

    for i in 1:2^T
        for j in 1:nλ
            P[i, j] = dynamic_probit(@view(Y[:, i]), β, λ[j])
            P[i, nλ + j] = dynamic_probit(@view(Y[:, 2^T + i]), β, λ[j])
        end
    end

end

# Format matrices for linear programming solution
function lp_format!(mod, p_test, β, hist_index)

    @unpack A, b, P, nx, nλ, λ, Y, T = mod

    # format matrices for LP
    A .= P' .- p_test'
    A[:, end] .= -1.0
    y0 = vcat(1, Y[:, hist_index])
    y1 = vcat(1, Y[:, 2^T + hist_index])
    for i in 1:nλ
        b[i] = dynamic_probit(y0, β, λ[i])
        b[nλ + i] = dynamic_probit(y1, β, λ[i])
    end
    b ./= @view(P[hist_index, :])

end

# Compute limits of identified set of forecast probabilities
function LP_limits!(p_grid, mod, β_grid, p, hist_index)

    @unpack T = mod

    # loop
    for (i, β) in enumerate(β_grid)

        # update cell probabilities
        make_cell_probabilities!(mod, β)

        # update matrices for linear programming solution
        lp_format!(mod, p, β, hist_index)

        # solve linear program for upper limit
        m = Model(Cbc.Optimizer)
        set_silent(m)
        set_time_limit_sec(m, 10)
        @variable(m, x[1:2^T])
        @unpack A, b = mod
        @objective(m, Min, x[2^T])
        @constraint(m, con, A * x .<= -b)
        optimize!(m)
        ts = termination_status(m)
        if ts == MOI.OPTIMAL
            pu = objective_value(m)
        end
        delete(m, con)

        if ts == MOI.OPTIMAL

            p_grid[i, 2] = pu

            # solve linear program for lower limit
            @objective(m, Max, -x[2^T])
            @constraint(m, A * x .<= b)
            optimize!(m)

            if termination_status(m) == MOI.OPTIMAL
                p_grid[i, 1] = objective_value(m)
            else
                p_grid[i, 1] = 1e6
            end

        else

            p_grid[i, 2] = -1e6
            p_grid[i, 1] = 1e6

        end
    end

end
